
#include <stdio.h>
#include <scale/btree.h>
#include <scale/alloc.h>


static sbtree_node_t *
_s_node_alloc(spointer key, spointer contents)
{
	sbtree_node_t *node;

	node = typed_malloc(sbtree_node_t);
	node->key = key;
	node->contents = contents;
	node->lnode = node->rnode = NULL;

	return node;
}


sbtree_t *
s_btree_alloc(sc_compf_t compare)
{
	sbtree_t *btree;

	btree = typed_malloc(sbtree_t);
	btree->count = 0;
	btree->compare = compare;
	btree->head = _s_node_alloc(NULL, NULL);
	btree->end = _s_node_alloc(NULL, NULL);

	btree->head->lnode = btree->head->rnode = btree->end;
	btree->end->lnode = btree->end->rnode = btree->end;

	return btree;
}


static void
_s_btree_node_free(sbtree_node_t *node, sbtree_node_t *end)
{
	if (node == end)
		return;
		
	_s_btree_node_free(node->lnode, end);
	_s_btree_node_free(node->rnode, end);
	s_freen(node, sizeof(sbtree_node_t));
}


void	
s_btree_free(sbtree_t *btree)
{
	_s_btree_node_free(btree->head, btree->end);
	s_freen(btree->end, sizeof(sbtree_node_t));
	s_freen(btree, sizeof(sbtree_t));
}


size_t	
s_btree_size(const sbtree_t *btree)
{
	return btree->count;
}


void	
s_btree_add(sbtree_t *btree, spointer key, spointer conts)
{
	sbtree_node_t *cnode, *pnode; /* child and parent node */
	sboolean save_right = true;
	
	pnode = btree->head;
	cnode = btree->head->rnode;
	
	while (cnode != btree->end) {
		pnode = cnode;
		save_right = (btree->compare(key, pnode->key) > 0);
		cnode = (save_right)?pnode->rnode:pnode->lnode;
	}

	cnode = _s_node_alloc(key, conts);
	cnode->lnode = cnode->rnode = btree->end;

	if (save_right)
		pnode->rnode = cnode;
	else
		pnode->lnode = cnode;

	++btree->count;
}


spointer	
s_btree_get(const sbtree_t *btree, sconstpointer key)
{
	sint cmp;
	sbtree_node_t *node;

	btree->end->key = (spointer)key;
	node = btree->head->rnode;
	
	while ((cmp = btree->compare(key, node->key)) != 0) {
		node = (cmp > 0)?node->rnode:node->lnode;
	}

	return node->contents;
}


spointer	
s_btree_remove(sbtree_t *btree, sconstpointer key)
{
	sint cmp;
	sbtree_node_t *pnode, *cnode, *tnode;
	spointer contents;
	sboolean save_right;

	btree->end->key = (spointer)key;
	pnode = btree->head;
	cnode = btree->head->rnode;
	save_right = true;

	while ((cmp = btree->compare(key, cnode->key)) != 0) {
		pnode = cnode;
		cnode = (cmp > 0)?cnode->rnode:cnode->lnode;
		save_right = (cmp > 0);
	}

	contents = cnode->contents;
	tnode = cnode;

	/* $B;R$r;}$?$J$$$+!":8$N;R$N$_;}$D$H$-$O(Bcnode$B$O:8$N;R$+(Bend$B$r;}$D(B */
	if (tnode->rnode == btree->end) {
		cnode = tnode->lnode;
	}
	else if (tnode->lnode == btree->end) {
		cnode = tnode->rnode;
	}
	/* $B1&$N;R$,:8$N;R$r;}$?$J$$;~$O!"(B */
	else if (tnode->rnode->lnode == btree->end) {
		cnode = tnode->rnode;
		cnode->lnode = tnode->lnode;
	}
	else {
		sbtree_node_t *nextnode; /* smallest node in 'cnode'. */
		nextnode = cnode->rnode;
		while (nextnode->lnode->lnode != btree->end)
			nextnode = nextnode->lnode;

		cnode = nextnode->lnode;
		nextnode->lnode = cnode->rnode;
		cnode->lnode = tnode->lnode;
		cnode->rnode = tnode->rnode;
	}

	s_freen(tnode, sizeof(sbtree_node_t));

	if (save_right)
		pnode->rnode = cnode;
	else
		pnode->lnode = cnode;

	--btree->count;
	return contents;
}


#include "scale/vector.h"
#include "scale/iterator.h"

/**
 * Left-First Search(lfs)$B$K$h$k%$%F%l!<%?9=B$BN!#(B
 * $BFsJ,LZ$G$N(Blfs$B$O$=$N9=B$>e!">:=g$KJB$V!#(B
 */
typedef struct {
	svector_t *nodes;
	size_t next_index;
} btree_lfs_itr_manager;



static spointer
_lfs_itr_next(siterator_t *itr)
{
	btree_lfs_itr_manager *mng;
	sbtree_node_t *node;
	
	mng = (btree_lfs_itr_manager *)itr->data;
	
	if (s_vector_size(mng->nodes) <= mng->next_index)
		return NULL;

	node = (sbtree_node_t *)s_vector_get(mng->nodes, mng->next_index++);

	return node->key;
}


static void
_lfs_itr_clear(siterator_t *itr)
{
	btree_lfs_itr_manager *mng;

	mng = (btree_lfs_itr_manager *)itr->data;

	mng->next_index = 0;
}

static void
_lfs_itr_free(siterator_t *itr)
{
	btree_lfs_itr_manager *mng;

	mng = (btree_lfs_itr_manager *)itr->data;

	s_vector_free(mng->nodes);
	s_freen(mng, sizeof(btree_lfs_itr_manager));
	s_freen(itr, sizeof(siterator_t));
}


static void
_lfs_add_node(svector_t *vect, sbtree_node_t *node, sbtree_node_t *end)
{
	if (node == end)
		return;

	_lfs_add_node(vect, node->lnode, end);
	s_vector_add(vect, node);
	_lfs_add_node(vect, node->rnode, end);
}

static svector_t *
_lfs_vector_alloc(const sbtree_t *btree)
{
	svector_t *nodes;

	nodes = s_vector_allocn(s_btree_size(btree));
	_lfs_add_node(nodes, btree->head->rnode, btree->end);
	return nodes;
}


siterator_t *
s_btree_lfs_iterator(const sbtree_t *btree)
{
	btree_lfs_itr_manager *mng;
	siterator_t *itr;

	mng = typed_malloc(btree_lfs_itr_manager);
	mng->nodes = _lfs_vector_alloc(btree);

	itr = typed_malloc(siterator_t);
	itr->data = mng;
	itr->next = _lfs_itr_next;
	itr->clear = _lfs_itr_clear;
	itr->free =  _lfs_itr_free;

	itr->clear(itr);

	return itr;
}


/* for sxtable_t */


#define GET_BTREE(xt) ((sbtree_t *)xt->tabledata)


static void
_s_btree_xtable_free(sxtable_t *xt)
{
	sbtree_t *btree = GET_BTREE(xt);

	if (xt->is_origin)
		s_btree_free(btree);
	s_freen(xt, sizeof(sxtable_t));
}


static size_t
_s_btree_xtable_size(const sxtable_t *xt)
{
	return s_btree_size(GET_BTREE(xt));
}


static void
_s_btree_xtable_put(sxtable_t *xt, spointer key, spointer contents)
{
	s_btree_add(GET_BTREE(xt), key, contents);
}


static spointer
_s_btree_xtable_get(const sxtable_t *xt, sconstpointer key)
{
	return s_btree_get(GET_BTREE(xt), key);
}


static void
_s_btree_xtable_remove(sxtable_t *xt, sconstpointer key)
{
	s_btree_remove(GET_BTREE(xt), key);
}


static siterator_t *
_s_btree_xtable_keys(const sxtable_t *xt)
{
	return s_btree_lfs_iterator(GET_BTREE(xt));
}


sxtable_t *
s_btree_attach_xtable(sbtree_t *btree)
{
	sxtable_t *xt;

	xt = typed_malloc(sxtable_t);
	xt->tabledata = btree;
	xt->is_origin = false;
	xt->free = _s_btree_xtable_free;
	xt->size = _s_btree_xtable_size;
	xt->put = _s_btree_xtable_put;
	xt->get = _s_btree_xtable_get;
	xt->remove = _s_btree_xtable_remove;
	xt->keys = _s_btree_xtable_keys;

	return xt;
}


sxtable_t *
s_btree_xtable(sc_compf_t cmpf)
{
	sxtable_t *xt;
	sbtree_t *btree;

	btree = s_btree_alloc(cmpf);
	xt = s_btree_attach_xtable(btree);
	xt->is_origin = true;

	return xt;
}
